@@ -1,6 +1,6 @@
 use v6;
 
-unit module TAP:ver<0.1.0>;
+unit module TAP:ver<0.3.5>;
 
 role Entry {
     has Str:D $.raw is required handles <Str>;
@@ -40,11 +40,9 @@ class Sub-Test is Test {
         my @plans = @!entries.grep(Plan);
         if !@plans {
             @errors.push: "Subtest $usable-number doesn't have a plan";
-        }
-        elsif @plans > 1 {
+        } elsif @plans > 1 {
             @errors.push: "Subtest $usable-number has multiple plans";
-        }
-        elsif @plans[0].tests != @tests.elems {
+        } elsif @plans[0].tests != @tests.elems {
             @errors.push: "Subtest $usable-number expected { @plans[0].tests } but contains { @tests.elems } tests";
         }
         @errors;
@@ -57,6 +55,9 @@ class Bailout does Entry {
 class Comment does Entry {
     has Str:D $.comment is required;
 }
+class Pragma does Entry {
+    has Bool:D %.identifiers is required;
+}
 class YAML does Entry {
     has Str:D $.serialized is required;
     has Any $.deserialized;
@@ -68,16 +69,9 @@ my role Entry::Handler {
     method handle-entry(Entry) { ... }
     method end-entries() { }
     method listen(Supply $supply) {
-        $supply.act(-> $entry {
-                self.handle-entry($entry);
-            },
-            done => {
-                self.end-entries();
-            },
-            quit => {
-                self.end-entries();
-            }
-        );
+        my $act = { self.handle-entry($^entry) };
+        my $done = my $quit = { self.end-entries() };
+        $supply.act($act, :$done, :$quit);
     }
 }
 
@@ -101,14 +95,14 @@ class Result {
         $!exit-status.defined ?? $!exit-status.exitcode !! Int;
     }
     method wait() {
-        $!exit-status.defined ?? $!exit-status.status !! Int;
+        $!exit-status.defined ?? ($!exit-status.exitcode +< 8) +| $!exit-status.signal !! Int;
     }
 
-    method has-problems($ignore-exit) {
-        @!failed || @!errors || (!$ignore-exit && self.exit-failed);
+    method has-problems($ignore-exit = False) {
+        so @!failed || @!errors || (!$ignore-exit && self.exit-failed);
     }
     method exit-failed() {
-        $!exit-status.defined && $!exit-status.status;
+        so $!exit-status.defined && ($!exit-status.exitcode || $!exit-status.signal);
     }
 }
 
@@ -157,10 +151,10 @@ class Aggregator {
 
 
     method has-problems() {
-        $!todo-passed || self.has-errors;
+        so $!todo-passed || self.has-errors;
     }
     method has-errors() {
-        $!failed + $!errors + $!exit-failed;
+        so $!failed + $!errors + $!exit-failed;
     }
     method get-status() {
         self.has-errors || $!tests-run != $!passed ?? 'FAILED' !! $!tests-run ?? 'PASS' !! 'NOTESTS';
@@ -169,7 +163,7 @@ class Aggregator {
 
 my grammar Grammar {
     regex TOP {
-        ^ [ <plan> | <test> | <bailout> | <version> | <comment> || <unknown> ] $
+        ^ [ <plan> | <test> | <bailout> | <version> | <pragma> | <comment> || <unknown> ] $
     }
     token sp { <[\s] - [\n]> }
     token num { <[0..9]>+ }
@@ -197,6 +191,12 @@ my grammar Grammar {
     token version {
         :i 'TAP version ' <version=.num>
     }
+    token pragma-identifier {
+        $<sign>=<[+-]> $<name>=[<alnum>+]
+    }
+    token pragma {
+        'pragma ' <pragma-identifier>+ % ' '
+    }
     token comment {
         '#' <.sp>* $<comment>=[\N*]
     }
@@ -206,7 +206,7 @@ my grammar Grammar {
         <.indent($indent)> '  ...'
     }
     token sub-entry(Int $indent) {
-        <plan> | <test> | <comment> | <yaml($indent)> | <sub-test($indent)> || <!before <.sp> > <unknown>
+        <plan> | <test> | <comment> | <pragma> | <yaml($indent)> | <sub-test($indent)> || <!before <.sp> > <unknown>
     }
     token indent(Int $indent) {
         '    ' ** { $indent }
@@ -250,6 +250,13 @@ my grammar Grammar {
         method version($/) {
             make TAP::Version.new(:raw(~$/), :version(+$<version>));
         }
+        method pragma-identifier($/) {
+            make $<name> => ?( $<sign> eq '+' );
+        }
+        method pragma($/) {
+            my Bool:D %identifiers = @<pragma-identifier>.map(*.ast);
+            make TAP::Pragma.new(:raw(~$/), :%identifiers);
+        }
         method comment($/) {
             make TAP::Comment.new(:raw(~$/), :comment(~$<comment>));
         }
@@ -309,29 +316,24 @@ my sub parser(Supply $input --> Supply) {
             if $mode == Normal {
                 if $line ~~ / ^ '  ---' / {
                     set-state(Yaml, $line);
-                }
-                elsif $line ~~ &indented {
+                } elsif $line ~~ &indented {
                     set-state(SubTest, $line);
-                }
-                else {
+                } else {
                     emit-reset $grammar.parse($line);
                 }
             }
             elsif $mode == SubTest {
                 if $line ~~ &indented {
                     @buffer.push: $line;
-                }
-                elsif $grammar.parse($line, :rule('test')) -> $test {
+                } elsif $grammar.parse($line, :rule('test')) -> $test {
                     my $raw = (|@buffer, $line).join("\n");
                     if $grammar.parse($raw, :rule('sub-test')) -> $subtest {
                         emit-reset $subtest;
-                    }
-                    else {
+                    } else {
                         emit-unknown;
                         emit-reset $test;
                     }
-                }
-                else {
+                } else {
                     emit-unknown $line;
                 }
             }
@@ -342,13 +344,11 @@ my sub parser(Supply $input --> Supply) {
                         my $raw = @buffer.join("\n");
                         if $grammar.parse($raw, :rule('yaml')) -> $yaml {
                             emit-reset $yaml;
-                        }
-                        else {
+                        } else {
                             emit-unknown;
                         }
                     }
-                }
-                else {
+                } else {
                     emit-unknown $line;
                 }
             }
@@ -357,6 +357,31 @@ my sub parser(Supply $input --> Supply) {
     }
 }
 
+role Output {
+    method print(Str $value) {
+        ...
+    }
+    method say(Str $value) {
+        self.print($value ~ "\n");
+    }
+    method flush() {
+    }
+    method terminal() {
+        False;
+    }
+}
+
+class Output::Handle does Output {
+    has IO::Handle:D $.handle handles(:print<print>, :flush<flush>, :terminal<t>) = $*OUT;
+}
+
+class Output::Supplier does Output {
+    has Supplier:D $.supplier is required;
+    method print(Str $value) {
+        $!supplier.emit($value);
+    }
+}
+
 enum Formatter::Volume <Silent Quiet Normal Verbose>;
 role Formatter {
     has Bool:D $.timer = False;
@@ -364,6 +389,8 @@ role Formatter {
     has Bool:D $.ignore-exit = False;
 }
 role Reporter {
+    has Output:D $.output is required;
+    has Formatter:D $.formatter handles<volume> is required;
     method summarize(TAP::Aggregator, Bool $interrupted, Duration $duration) { ... }
     method open-test(Str $) { ... }
 }
@@ -425,8 +452,7 @@ class Formatter::Text does Formatter {
                     if $result.wait -> $wait {
                         if $result.exit {
                             $output ~= self.format-failure("Non-zero exit status: { $result.exit }\n");
-                        }
-                        else {
+                        } else {
                             $output ~= self.format-failure("Non-zero wait status: $wait\n");
                         }
                     }
@@ -457,11 +483,9 @@ class Formatter::Text does Formatter {
         my $name = $session.header;
         if ($result.skip-all) {
             $output = self.format-return("$name skipped\n");
-        }
-        elsif ($result.has-problems($!ignore-exit)) {
+        } elsif ($result.has-problems($!ignore-exit)) {
             $output = self.format-test-failure($name, $result);
-        }
-        else {
+        } else {
             my $time = self.timer && $result.time ?? sprintf ' %8d ms', Int($result.time * 1000) !! '';
             $output = self.format-return("$name ok$time\n");
         }
@@ -480,8 +504,7 @@ class Formatter::Text does Formatter {
 
         if $result.failed == 0 {
             $output ~= self.format-failure($total ?? "All $total subtests passed " !! 'No subtests run');
-        }
-        else {
+        } else {
             $output ~= self.format-failure("Failed {$result.failed.elems}/$total subtests ");
             if (!$total) {
                 $output ~= self.format-failure("\nNo tests run!");
@@ -504,13 +527,6 @@ class Formatter::Text does Formatter {
     }
 }
 class Reporter::Text does Reporter {
-    has IO::Handle $!handle;
-    has Formatter::Text $!formatter handles <volume>;
-
-    submethod BUILD(:@names, :$!handle = $*OUT, :$volume = Normal, :$timer = False, Bool :$ignore-exit) {
-        $!formatter = Formatter::Text.new(:@names, :$volume, :$timer, :$ignore-exit);
-    }
-
     method open-test(Str $name) {
         my $header = $!formatter.format-name($name);
         Reporter::Text::Session.new(:$name, :$header, :reporter(self));
@@ -519,16 +535,15 @@ class Reporter::Text does Reporter {
         self!output($!formatter.format-summary($aggregator, $interrupted, $duration)) unless self.volume === Silent;
     }
     method !output(Any $value) {
-        $!handle.print($value);
+        $!output.print($value);
     }
     method print-result(Reporter::Text::Session $session, TAP::Result $report) {
         self!output($!formatter.format-result($session, $report)) unless self.volume <= Quiet;
     }
 }
 
-class Formatter::Console is Formatter::Text {
-    has $.color;
-    has &colored = $!color && (try require Terminal::ANSIColor) !=== Nil
+class Formatter::Color is Formatter::Text {
+    has &colored = (try require Terminal::ANSIColor) !=== Nil
         ?? ::('Terminal::ANSIColor::EXPORT::DEFAULT::&colored')
         !! sub ($text, $) { $text };
     method format-success(Str $output) {
@@ -571,20 +586,13 @@ class Reporter::Console::Session does Session {
     }
 }
 class Reporter::Console does Reporter {
-    has Formatter::Console $!formatter handles <volume>;
-    has Int $!lastlength;
-    has Supplier $events;
+    has Int $!lastlength = 0;
+    has Supplier $events = Supplier.new;
     has Reporter::Console::Session @!active;
-    has Int $!tests;
-    has Int $!fails;
-
-    submethod BUILD(:@names, IO::Handle :$handle = $*OUT, :$volume = Normal, :$timer = False, Bool :$ignore-exit = False, Bool :$color) {
-        $!formatter = Formatter::Console.new(:@names, :$volume, :$timer, :$ignore-exit, :$color);
-        $!lastlength = 0;
-        $!events = Supplier.new;
-        @!active .= new;
-        $!tests = 0;
+    has Int $!tests = 0;
+    has Int $!fails = 0;
 
+    submethod TWEAK(:$!output, :$!formatter) {
         my $now = 0;
         my $start = now;
 
@@ -596,31 +604,30 @@ class Reporter::Console does Reporter {
             my @items = @!active.map(*.summary);
             my $ruler = ($header, |@items).join('  ') ~ ')===';
             $ruler = $ruler.substr(0,70) if $ruler.chars > 70;
-            $handle.print($!formatter.format-return($ruler));
+            $!output.print($!formatter.format-return($ruler));
         }
         multi receive('update', Str $name, Str $header, Int $number, Int $plan) {
             return if self.volume <= Quiet;
             if @!active.elems == 1 {
                 my $status = ($header, $number, '/', $plan // '?').join('');
-                $handle.print($!formatter.format-return($status));
+                $!output.print($!formatter.format-return($status));
                 $!lastlength = $status.chars + 1;
-            }
-            else {
+            } else {
                 output-ruler($number == 1);
             }
         }
         multi receive('bailout', Str $explanation) {
             return if self.volume <= Quiet;
-            $handle.print($!formatter.format-failure("Bailout called.  Further testing stopped: $explanation\n"));
+            $!output.print($!formatter.format-failure("Bailout called.  Further testing stopped: $explanation\n"));
         }
         multi receive('result', Reporter::Console::Session $session, TAP::Result $result) {
             return if self.volume <= Quiet;
-            $handle.print($!formatter.format-return(' ' x $!lastlength) ~ $!formatter.format-result($session, $result));
+            $!output.print($!formatter.format-return(' ' x $!lastlength) ~ $!formatter.format-result($session, $result));
             @!active = @!active.grep(* !=== $session);
             output-ruler(True) if @!active.elems > 1;
         }
         multi receive('summary', TAP::Aggregator $aggregator, Bool $interrupted, Duration $duration) {
-            $handle.print($!formatter.format-summary($aggregator, $interrupted, $duration)) unless self.volume == Silent;
+            $!output.print($!formatter.format-summary($aggregator, $interrupted, $duration)) unless self.volume == Silent;
         }
 
         $!events.Supply.act(-> @args { receive(|@args) });
@@ -648,7 +655,7 @@ class Reporter::Console does Reporter {
 }
 
 my class State does TAP::Entry::Handler {
-    has Range $.allowed-versions = 12 .. 13;
+    has Range $.allowed-versions = 12 .. 14;
     has Int $!tests-planned;
     has Int $!tests-run = 0;
     has Int $!passed = 0;
@@ -661,6 +668,7 @@ my class State does TAP::Entry::Handler {
     has Int $!skipped = 0;
     has Int $!unknowns = 0;
     has Bool $!skip-all = False;
+    has Bool $!strict;
 
     has Promise $.bailout;
     has Int $!seen-lines = 0;
@@ -680,19 +688,16 @@ my class State does TAP::Entry::Handler {
     multi method handle-entry(TAP::Version $entry) {
         if $!seen-lines {
             self!add-error('Seen version declaration mid-stream');
-        }
-        elsif $entry.version !~~ $!allowed-versions {
+        } elsif $entry.version !~~ $!allowed-versions {
             self!add-error("Version must be in range $!allowed-versions");
-        }
-        else {
+        } else {
             $!version = $entry.version;
         }
     }
     multi method handle-entry(TAP::Plan $plan) {
         if $!seen-plan != Unseen {
             self!add-error('Seen a second plan');
-        }
-        else {
+        } else {
             $!tests-planned = $plan.tests;
             $!seen-plan = $!tests-run ?? After !! Before;
             $!skip-all = $plan.skip-all;
@@ -711,8 +716,7 @@ my class State does TAP::Entry::Handler {
         my $usable-number = $found-number // $expected-number;
         if $test.is-ok {
             $!passed++;
-        }
-        else {
+        } else {
             @!failed.push($usable-number);
         }
         ($test.ok ?? $!actual-passed !! $!actual-failed)++;
@@ -729,13 +733,20 @@ my class State does TAP::Entry::Handler {
     multi method handle-entry(TAP::Bailout $entry) {
         if $!bailout.defined {
             $!bailout.break($entry);
-        }
-        else {
+        } else {
             $!done.break($entry);
         }
     }
     multi method handle-entry(TAP::Unknown $) {
         $!unknowns++;
+        if $!strict {
+            self!add-error('Seen invalid TAP in strict mode');
+        }
+    }
+    multi method handle-entry(TAP::Pragma $entry) {
+        if $entry.identifiers<strict>:exists {
+            $!strict = $entry.identifiers<strict>;
+        }
     }
     multi method handle-entry(TAP::Entry $entry) {
     }
@@ -783,9 +794,10 @@ class Source::Proc does Source {
     has Str $.path is required;
     has @.args;
     has $.err is required;
+    has $.cwd is required;
 }
 class Source::File does Source {
-    has Str $.filename;
+    has IO(Str) $.filename handles<slurp>;
 }
 class Source::String does Source {
     has Str $.content;
@@ -823,16 +835,16 @@ class Async {
                 $async.bind-stderr($devnull);
             }
             when IO::Handle:D {
-                $async.stderr.lines(:close).act({ $err.say($_) });
+                $async.bind-stderr($err);
             }
-            when Supply:D {
+            when Supplier:D {
                 $async.stderr.act({ $err.emit($_) }, :done({ $err.done }), :quit({ $err.quit($^reason) }));
             }
             default {
                 die "Unknown error handler";
             }
         }
-        my $process = $async.start;
+        my $process = $async.start(:cwd($proc.cwd));
         my $start-time = now;
         my $timer = $process.then({ now - $start-time });
         Run.new(:$process, :killer($async), :$timer, :$events);
@@ -843,7 +855,7 @@ class Async {
         Run.new(:$events);
     }
     multi get_runner(Source::File $file) {
-        my $events = parser(supply { emit $file.filename.IO.slurp(:close) });
+        my $events = parser(supply { emit $file.slurp(:close) });
         Run.new(:$events);
     }
     multi get_runner(Source::String $string) {
@@ -871,15 +883,17 @@ class Harness {
         method make-source {...};
     }
     role SourceHandler::Proc does SourceHandler {
-        has Str:D $.path is required;
-        has @.args;
-        method make-source(Str:D $name, Any:D :$err) {
-            TAP::Source::Proc.new(:$name, :$!path, :args[ |@!args, $name ], :$err);
-        }
+    }
+    my sub normalize-path($path, IO::Path $cwd) {
+        $path ~~ IO ?? $path.IO.relative($cwd) !! ~$path
     }
     class SourceHandler::Raku does SourceHandler::Proc {
-        submethod BUILD(:@incdirs, Str:D :$!path = $*EXECUTABLE.absolute) {
-            @!args = @incdirs.map("-I" ~ *);
+        has Str:D $.path = $*EXECUTABLE.absolute;
+        has @.incdirs;
+        method make-source(Str:D $name, Any:D :$err, IO::Path:D :$cwd, :@include-dirs = (), *%) {
+            my @dirs = map { normalize-path($_, $cwd) }, |@include-dirs, |@!incdirs;
+            my @args = |@dirs.map("-I" ~ *), $name;
+            TAP::Source::Proc.new(:$name, :$!path, :@args, :$err, :$cwd);
         }
         method can-handle(Str $name) {
             0.5;
@@ -889,12 +903,18 @@ class Harness {
         # This may later give deprecation warnings
     }
     class SourceHandler::Exec does SourceHandler::Proc {
+        has Str:D $.path is required;
+        has @.args;
         method new (*@ ($path, *@args)) {
             self.bless(:$path, :@args);
         }
         method can-handle(Str $name) {
             1;
         }
+        method make-source(Str:D $name, Any:D :$err, IO::Path:D :$cwd, *%) {
+            my @args = |@!args, $name;
+            TAP::Source::Proc.new(:$name, :$!path, :@args, :$err, :$cwd);
+        }
     }
 
     my sub load-reporter(Str $name --> Reporter) {
@@ -905,67 +925,108 @@ class Harness {
         return ::($classname);
     }
 
+    subset OutVal where any(IO::Handle:D, Supplier:D);
+
     has SourceHandler @.handlers = SourceHandler::Raku.new();
     has IO::Handle $.handle = $*OUT;
+    has OutVal $.output = $!handle;
     has Formatter::Volume $.volume = ?%*ENV<HARNESS_VERBOSE> ?? Verbose !! Normal;
     has Str %!env-options = (%*ENV<HARNESS_OPTIONS> // '').split(':').grep(*.chars).map: { / ^ (.) (.*) $ /; ~$0 => val(~$1) };
-    has TAP::Reporter:U $.reporter-class = %!env-options<r> ?? load-reporter(%!env-options<r>) !! $!handle.t && $!volume < Verbose ?? TAP::Reporter::Console !! TAP::Reporter::Text;
+    has TAP::Reporter:U $.reporter-class;
     has Int:D $.jobs = %!env-options<j> // 1;
     has Bool:D $.timer = ?%*ENV<HARNESS_TIMER>;
-    subset ErrValue where any(IO::Handle:D, Supply, 'stderr', 'ignore', 'merge');
+    subset ErrValue where any(IO::Handle:D, Supplier, 'stderr', 'ignore', 'merge');
     has ErrValue $.err = 'stderr';
     has Bool:D $.ignore-exit = ?%*ENV<HARNESS_INGORE_EXIT>;
     has Bool:D $.trap = False;
     has Bool:D $.loose = $*PERL.compiler.version before 2017.09;
-    my @safe-terminals = <xterm eterm vte konsole color>;
-    has Bool:D $.color = so %!env-options<c> // %*ENV<HARNESS_COLOR> // !%*ENV<NO_COLOR>.defined && $!handle.t && (%*ENV<TERM> // '') ~~ / @safe-terminals /;
+    has Bool $.color;
 
-    class Run {
-        has Promise $.waiter handles <result>;
-        has Promise $!killed;
-        submethod BUILD (Promise :$!waiter, Promise :$!killed) {
+    class Run does Awaitable {
+        has Promise $!waiter handles <result>;
+        has Promise $!bailout;
+        submethod BUILD (Promise :$!waiter, Promise :$!bailout) {
         }
         method kill(Any:D $reason) {
-            $!killed.break($reason);
+            $!bailout.break($reason);
+        }
+        method get-await-handle(--> Awaitable::Handle:D) {
+            return $!waiter.promise.get-await-handle;
         }
     }
 
     method make-aggregator() {
         TAP::Aggregator.new(:$!ignore-exit);
     }
-    method add-handlers(Supply $events) {
+    method add-handlers(Supply $events, Output $output) {
         if $.volume == Verbose {
-            $events.act({ $!handle.say(~$^entry) }, :done({ $!handle.flush }), :quit({ $!handle.flush }));
+            $events.act({ $output.say(~$^entry) }, :done({ $output.flush }), :quit({ $output.flush }));
         }
     }
-    method make-source(Str $name) {
-        @!handlers.max(*.can-handle($name)).make-source($name, :$!err);
-    }
     my &sigint = sub { signal(SIGINT) }
 
-    method run(*@sources) {
-        my $killed = Promise.new;
+    my multi make-output(IO::Handle:D $handle) {
+        return Output::Handle.new(:$handle);
+    }
+    my multi make-output(Supplier:D $supplier) {
+        return Output::Supplier.new(:$supplier);
+    }
+
+    my sub get-reporter(Reporter $reporter-class, %env-options, $output, $volume) {
+        if $reporter-class !=== Reporter {
+            return $reporter-class;
+        } elsif %env-options<r> {
+            return load-reporter(%env-options<r>);
+        } elsif $output.terminal && $volume < Verbose {
+            return TAP::Reporter::Console;
+        } else {
+            return TAP::Reporter::Text;
+        }
+    }
+
+    my sub get-color(Bool $color, %env-options, Output $output) {
+        with $color {
+            return $color;
+        } orwith %env-options<c> {
+            return True;
+        } orwith %*ENV<HARNESS_COLOR> {
+            return ?%*ENV<HARNESS_COLOR>;
+        } orwith %*ENV<NO_COLOR> {
+            return False;
+        } else {
+            state @safe-terminals = <xterm eterm vte konsole color>;
+            return $output.terminal && (%*ENV<TERM> // '') ~~ / @safe-terminals /;
+        }
+    }
+
+    method run(*@names, IO(Str) :$cwd = $*CWD, OutVal :$out = $!output, ErrValue :$err = $!err, *%handler-args) {
+        my $bailout = Promise.new;
         my $aggregator = self.make-aggregator;
-        my $reporter = $!reporter-class.new(:names(@sources), :$!timer, :$!ignore-exit, :$!volume, :$!handle, :$!color);
+        my $output = make-output($out);
+        my $formatter-class = get-color($!color, %!env-options, $output) ?? Formatter::Color !! Formatter::Text;
+        my $formatter = $formatter-class.new(:@names, :$!volume, :$!timer, :$!ignore-exit);
+        my $reporter-class = get-reporter($!reporter-class, %!env-options, $output, $!volume);
+        my $reporter = $reporter-class.new(:$output, :$formatter);
 
         my @working;
         my $waiter = start {
-            my $int = $!trap ?? sigint().tap({ $killed.break("Interrupted"); $int.close(); }) !! Tap;
+            my $int = $!trap ?? sigint().tap({ $bailout.break("Interrupted"); $int.close(); }) !! Tap;
             my $begin = now;
             try {
-                for @sources -> $name {
-                    my $session = $reporter.open-test($name);
-                    my $source = self.make-source($name);
-                    my $parser = TAP::Async.new(:$source, :$killed, :$!loose);
+                for @names -> $name {
+                    my $path = normalize-path($name, $cwd);
+                    my $session = $reporter.open-test($path);
+                    my $source = @!handlers.max(*.can-handle($name)).make-source($path, :$err, :$cwd, |%handler-args);
+                    my $parser = TAP::Async.new(:$source, :$bailout, :$!loose);
                     $session.listen($parser.events);
-                    self.add-handlers($parser.events);
+                    self.add-handlers($parser.events, $output);
                     @working.push({ :$parser, :$session, :done($parser.waiter) });
                     next if @working < $!jobs;
-                    await Promise.anyof(@working»<done>, $killed);
+                    await Promise.anyof(@working»<done>, $bailout);
                     reap-finished();
                 }
                 while @working {
-                    await Promise.anyof(@working»<done>, $killed);
+                    await Promise.anyof(@working»<done>, $bailout);
                     reap-finished();
                 }
                 CATCH {
@@ -975,24 +1036,23 @@ class Harness {
                     }
                 }
             }
-            $reporter.summarize($aggregator, ?$killed, now - $begin) if !$killed || $!trap;
+            $reporter.summarize($aggregator, ?$bailout, now - $begin) if !$bailout || $!trap;
             $int.close if $int;
             $aggregator;
         }
         sub reap-finished() {
             my @new-working;
-            for @working -> $current {
-                if $current<done> {
-                    $aggregator.add-result($current<parser>.result);
-                    $current<session>.close-test($current<parser>.result);
-                }
-                else {
+            for @working -> $current (:$done, :$parser, :$session) {
+                if $done {
+                    $aggregator.add-result($parser.result);
+                    $session.close-test($parser.result);
+                } else {
                     @new-working.push($current);
                 }
             }
             @working = @new-working;
         }
-        Run.new(:$waiter, :$killed);
+        Run.new(:$waiter, :$bailout);
     }
 }
 
